@@ -4,6 +4,7 @@
  * Copyright (C) 1999 Red Hat Software
  * Copyright (C) 2000 Tor Lillqvist
  * Copyright (C) 2001 Alexander Larsson
+ * Copyright (C) 2007 Novell, Inc.
  *
  * This library is free software; you can redistribute it and/or
  * modify it under the terms of the GNU Library General Public
@@ -31,14 +32,16 @@
 #include "pangowin32.h"
 #include "pangowin32-private.h"
 
+#define MAX_FREED_FONTS 16
+
 #define CH_IS_UNIHAN_BMP(ch) ((ch) >= 0x3400 && (ch) <= 0x9FFF)
 #define CH_IS_UNIHAN(ch) (CH_IS_UNIHAN_BMP (ch) || \
 			  ((ch) >= 0x20000 && (ch) <= 0x2A6DF) ||	\
 			  ((ch) >= 0x2F800 && (ch) <= 0x2FA1F))
 
-HDC pango_win32_hdc;
-OSVERSIONINFO pango_win32_os_version_info;
-gboolean pango_win32_debug = FALSE;
+HDC _pango_win32_hdc;
+OSVERSIONINFO _pango_win32_os_version_info;
+gboolean _pango_win32_debug = FALSE;
 
 static void pango_win32_font_dispose    (GObject             *object);
 static void pango_win32_font_finalize   (GObject             *object);
@@ -92,33 +95,33 @@ pango_win32_get_hfont (PangoFont *font)
   if (!win32font->hfont)
     {
       cache = pango_win32_font_map_get_font_cache (win32font->fontmap);
-      
-      win32font->hfont = pango_win32_font_cache_load (cache, &win32font->logfont);
+
+      win32font->hfont = pango_win32_font_cache_loadw (cache, &win32font->logfontw);
       if (!win32font->hfont)
 	{
-	  gchar *face_utf8 = g_locale_to_utf8 (win32font->logfont.lfFaceName,
-					       -1, NULL, NULL, NULL);
+	  gchar *face_utf8 = g_utf16_to_utf8 (win32font->logfontw.lfFaceName,
+					      -1, NULL, NULL, NULL);
 	  g_warning ("Cannot load font '%s\n", face_utf8);
 	  g_free (face_utf8);
 	  return NULL;
 	}
 
-      SelectObject (pango_win32_hdc, win32font->hfont);
-      GetTextMetrics (pango_win32_hdc, &tm);
+      SelectObject (_pango_win32_hdc, win32font->hfont);
+      GetTextMetrics (_pango_win32_hdc, &tm);
 
       win32font->tm_overhang = tm.tmOverhang;
       win32font->tm_descent = tm.tmDescent;
       win32font->tm_ascent = tm.tmAscent;
     }
-  
+
   return win32font->hfont;
 }
 
 /**
  * pango_win32_get_context:
- * 
+ *
  * Retrieves a #PangoContext appropriate for rendering with Windows fonts.
-  * 
+  *
  * Return value: the new #PangoContext
  **/
 PangoContext *
@@ -132,10 +135,10 @@ pango_win32_get_context (void)
   return result;
 }
 
-G_DEFINE_TYPE (PangoWin32Font, pango_win32_font, PANGO_TYPE_FONT)
+G_DEFINE_TYPE (PangoWin32Font, _pango_win32_font, PANGO_TYPE_FONT)
 
-static void 
-pango_win32_font_init (PangoWin32Font *win32font)
+static void
+_pango_win32_font_init (PangoWin32Font *win32font)
 {
   win32font->size = -1;
 
@@ -146,22 +149,22 @@ pango_win32_font_init (PangoWin32Font *win32font)
 
 /**
  * pango_win32_get_dc:
- * 
+ *
  * Obtains a handle to the Windows device context that is used by Pango.
- * 
+ *
  * Return value: A handle to the Windows device context that is used by Pango.
  **/
 HDC
 pango_win32_get_dc (void)
 {
-  if (pango_win32_hdc == NULL)
+  if (_pango_win32_hdc == NULL)
     {
-      pango_win32_hdc = CreateDC ("DISPLAY", NULL, NULL, NULL);
-      memset (&pango_win32_os_version_info, 0,
-	      sizeof (pango_win32_os_version_info));
-      pango_win32_os_version_info.dwOSVersionInfoSize =
+      _pango_win32_hdc = CreateDC ("DISPLAY", NULL, NULL, NULL);
+      memset (&_pango_win32_os_version_info, 0,
+	      sizeof (_pango_win32_os_version_info));
+      _pango_win32_os_version_info.dwOSVersionInfoSize =
 	sizeof (OSVERSIONINFO);
-      GetVersionEx (&pango_win32_os_version_info);
+      GetVersionEx (&_pango_win32_os_version_info);
 
       /* Also do some generic pangowin32 initialisations... this function
        * is a suitable place for those as it is called from a couple
@@ -169,37 +172,37 @@ pango_win32_get_dc (void)
        */
 #ifdef PANGO_WIN32_DEBUGGING
       if (getenv ("PANGO_WIN32_DEBUG") != NULL)
-	pango_win32_debug = TRUE;
-#endif      
+	_pango_win32_debug = TRUE;
+#endif
     }
 
-  return pango_win32_hdc;
-}  
+  return _pango_win32_hdc;
+}
 
 /**
  * pango_win32_get_debug_flag:
  *
  * Returns whether debugging is turned on.
- * 
+ *
  * Return value: %TRUE if debugging is turned on.
- * 
+ *
  * Since: 1.2
  */
 gboolean
 pango_win32_get_debug_flag (void)
 {
-  return pango_win32_debug;
+  return _pango_win32_debug;
 }
 
 static void
-pango_win32_font_class_init (PangoWin32FontClass *class)
+_pango_win32_font_class_init (PangoWin32FontClass *class)
 {
   GObjectClass *object_class = G_OBJECT_CLASS (class);
   PangoFontClass *font_class = PANGO_FONT_CLASS (class);
 
   object_class->finalize = pango_win32_font_finalize;
   object_class->dispose = pango_win32_font_dispose;
-  
+
   font_class->describe = pango_win32_font_describe;
   font_class->describe_absolute = pango_win32_font_describe_absolute;
   font_class->get_coverage = pango_win32_font_get_coverage;
@@ -215,28 +218,6 @@ pango_win32_font_class_init (PangoWin32FontClass *class)
   pango_win32_get_dc ();
 }
 
-PangoWin32Font *
-pango_win32_font_new (PangoFontMap  *fontmap,
-		      const LOGFONT *lfp,
-		      int	     size)
-{
-  PangoWin32Font *result;
-
-  g_return_val_if_fail (fontmap != NULL, NULL);
-  g_return_val_if_fail (lfp != NULL, NULL);
-
-  result = (PangoWin32Font *)g_object_new (PANGO_TYPE_WIN32_FONT, NULL);
-  
-  result->fontmap = fontmap;
-  g_object_ref (fontmap);
-
-  result->size = size;
-  pango_win32_make_matching_logfont (fontmap, lfp, size,
-				     &result->logfont);
-
-  return result;
-}
-
 /**
  * pango_win32_render:
  * @hdc:     the device context
@@ -247,11 +228,11 @@ pango_win32_font_new (PangoFontMap  *fontmap,
  *
  * Render a #PangoGlyphString onto a Windows DC
  */
-void 
+void
 pango_win32_render (HDC               hdc,
 		    PangoFont        *font,
 		    PangoGlyphString *glyphs,
-		    int               x, 
+		    int               x,
 		    int               y)
 {
   HFONT hfont, old_hfont = NULL;
@@ -264,13 +245,13 @@ pango_win32_render (HDC               hdc,
   g_return_if_fail (glyphs != NULL);
 
 #ifdef PANGO_WIN32_DEBUGGING
-  if (pango_win32_debug)
+  if (_pango_win32_debug)
     {
       PING (("num_glyphs:%d", glyphs->num_glyphs));
       for (i = 0; i < glyphs->num_glyphs; i++)
 	{
 	  g_print (" %d:%d", glyphs->glyphs[i].glyph, glyphs->glyphs[i].geometry.width);
-	  if (glyphs->glyphs[i].geometry.x_offset != 0 || 
+	  if (glyphs->glyphs[i].geometry.x_offset != 0 ||
 	      glyphs->glyphs[i].geometry.y_offset != 0)
 	    g_print (":%d,%d", glyphs->glyphs[i].geometry.x_offset,
 		     glyphs->glyphs[i].geometry.y_offset);
@@ -287,7 +268,7 @@ pango_win32_render (HDC               hdc,
     return;
 
   old_hfont = SelectObject (hdc, hfont);
-		
+
   glyph_indexes = g_new (guint16, glyphs->num_glyphs);
   dX = g_new (INT, glyphs->num_glyphs);
 
@@ -372,7 +353,7 @@ pango_win32_render (HDC               hdc,
 	  i++;
 	}
 #ifdef PANGO_WIN32_DEBUGGING
-      if (pango_win32_debug)
+      if (_pango_win32_debug)
 	{
 	  g_print ("ExtTextOutW at %d,%d deltas:",
 		   x + PANGO_PIXELS (start_x_offset),
@@ -417,12 +398,12 @@ pango_win32_render (HDC               hdc,
  * must have been loaded using a #PangoContext with an identical
  * transformation matrix to that passed in to this function.
  **/
-void 
+void
 pango_win32_render_transformed (HDC                hdc,
 				const PangoMatrix *matrix,
 				PangoFont         *font,
 				PangoGlyphString  *glyphs,
-				int                x, 
+				int                x,
 				int                y)
 {
   XFORM xForm;
@@ -442,7 +423,7 @@ pango_win32_render_transformed (HDC                hdc,
       xForm.eDx = matrix->x0;
       xForm.eDy = matrix->y0;
       if (!SetWorldTransform (hdc, &xForm))
-        g_warning ("GetWorldTransform() failed");
+	g_warning ("GetWorldTransform() failed");
     }
 
   pango_win32_render (hdc, font, glyphs, x/PANGO_SCALE, y/PANGO_SCALE);
@@ -479,7 +460,7 @@ pango_win32_font_get_glyph_extents (PangoFont      *font,
     glyph_index = glyph = 0;
 
   info = g_hash_table_lookup (win32font->glyph_info, GUINT_TO_POINTER (glyph));
-  
+
   if (!info)
     {
       info = g_new0 (PangoWin32GlyphInfo, 1);
@@ -487,16 +468,16 @@ pango_win32_font_get_glyph_extents (PangoFont      *font,
       memset (&gm, 0, sizeof (gm));
 
       hfont = pango_win32_get_hfont (font);
-      SelectObject (pango_win32_hdc, hfont);
-      /* FIXME: (Alex) This constant reuse of pango_win32_hdc is
+      SelectObject (_pango_win32_hdc, hfont);
+      /* FIXME: (Alex) This constant reuse of _pango_win32_hdc is
 	 not thread-safe */
-      res = GetGlyphOutlineA (pango_win32_hdc, 
+      res = GetGlyphOutlineA (_pango_win32_hdc,
 			      glyph_index,
 			      GGO_METRICS | GGO_GLYPH_INDEX,
-			      &gm, 
+			      &gm,
 			      0, NULL,
 			      &m);
-  
+
       if (res == GDI_ERROR)
 	{
 	  gchar *error = g_win32_error_message (GetLastError ());
@@ -533,19 +514,19 @@ max_glyph_width (PangoLayout *layout)
   int max_width = 0;
   GSList *l, *r;
 
-  for (l = pango_layout_get_lines (layout); l; l = l->next)
+  for (l = pango_layout_get_lines_readonly (layout); l; l = l->next)
     {
       PangoLayoutLine *line = l->data;
 
       for (r = line->runs; r; r = r->next)
-        {
-          PangoGlyphString *glyphs = ((PangoGlyphItem *)r->data)->glyphs;
-          int i;
+	{
+	  PangoGlyphString *glyphs = ((PangoGlyphItem *)r->data)->glyphs;
+	  int i;
 
-          for (i = 0; i < glyphs->num_glyphs; i++)
-            if (glyphs->glyphs[i].geometry.width > max_width)
-              max_width = glyphs->glyphs[i].geometry.width;
-        }
+	  for (i = 0; i < glyphs->num_glyphs; i++)
+	    if (glyphs->glyphs[i].geometry.width > max_width)
+	      max_width = glyphs->glyphs[i].geometry.width;
+	}
     }
 
   return max_width;
@@ -558,14 +539,14 @@ pango_win32_font_get_metrics (PangoFont     *font,
   PangoWin32MetricsInfo *info = NULL; /* Quiet gcc */
   PangoWin32Font *win32font = (PangoWin32Font *)font;
   GSList *tmp_list;
-      
+
   const char *sample_str = pango_language_get_sample_string (language);
-  
+
   tmp_list = win32font->metrics_by_lang;
   while (tmp_list)
     {
       info = tmp_list->data;
-      
+
       if (info->sample_str == sample_str)    /* We _don't_ need strcmp */
 	break;
 
@@ -576,7 +557,7 @@ pango_win32_font_get_metrics (PangoFont     *font,
     {
       HFONT hfont;
       PangoFontMetrics *metrics;
-      
+
       info = g_new (PangoWin32MetricsInfo, 1);
       win32font->metrics_by_lang = g_slist_prepend (win32font->metrics_by_lang, info);
 
@@ -588,9 +569,9 @@ pango_win32_font_get_metrics (PangoFont     *font,
 	{
 	  PangoCoverage *coverage;
 	  TEXTMETRIC tm;
-	    
-	  SelectObject (pango_win32_hdc, hfont);
-	  GetTextMetrics (pango_win32_hdc, &tm);
+
+	  SelectObject (_pango_win32_hdc, hfont);
+	  GetTextMetrics (_pango_win32_hdc, &tm);
 
 	  metrics->ascent = tm.tmAscent * PANGO_SCALE;
 	  metrics->descent = tm.tmDescent * PANGO_SCALE;
@@ -603,7 +584,7 @@ pango_win32_font_get_metrics (PangoFont     *font,
 	      PangoContext *context;
 	      PangoFontDescription *font_desc;
 	      PangoLayout *layout;
-	      
+
 	      /*  Get the average width of the chars in "0123456789" */
 	      context = pango_win32_get_context ();
 	      pango_context_set_language (context, language);
@@ -631,7 +612,7 @@ pango_win32_font_get_metrics (PangoFont     *font,
 	  metrics->strikethrough_position = metrics->ascent / 3;
 	}
     }
-      
+
   return pango_font_metrics_ref (info->metrics);
 }
 
@@ -651,13 +632,13 @@ pango_win32_font_real_select_font (PangoFont *font,
 
   if (!hfont)
     return FALSE;
-  
+
   if (!SelectObject (hdc, hfont))
     {
       g_warning ("pango_win32_font_real_select_font: Cannot select font\n");
       return FALSE;
     }
-  
+
   return TRUE;
 }
 
@@ -675,23 +656,61 @@ pango_win32_font_real_get_metrics_factor (PangoFont *font)
 /**
  * pango_win32_font_logfont:
  * @font: a #PangoFont which must be from the Win32 backend
- * 
- * Determine the LOGFONT struct for the specified font.
- * 
- * Return value: A newly allocated LOGFONT struct. It must be
+ *
+ * Determine the LOGFONTA struct for the specified font. Note that
+ * Pango internally uses LOGFONTW structs, so if converting the UTF-16
+ * face name in the LOGFONTW struct to system codepage fails, the
+ * returned LOGFONTA will have an emppty face name. To get the
+ * LOGFONTW of a PangoFont, use pango_win32_font_logfontw(). It
+ * is recommended to do that always even if you don't expect
+ * to come across fonts with odd names.
+ *
+ * Return value: A newly allocated LOGFONTA struct. It must be
  * freed with g_free().
  **/
-LOGFONT *
+LOGFONTA *
 pango_win32_font_logfont (PangoFont *font)
 {
   PangoWin32Font *win32font = (PangoWin32Font *)font;
-  LOGFONT *lfp;
+  LOGFONTA *lfp;
 
   g_return_val_if_fail (font != NULL, NULL);
   g_return_val_if_fail (PANGO_WIN32_IS_FONT (font), NULL);
 
-  lfp = g_new (LOGFONT, 1);
-  *lfp = win32font->logfont;
+  lfp = g_new (LOGFONTA, 1);
+
+  *lfp = *(LOGFONTA*) &win32font->logfontw;
+  if (!WideCharToMultiByte (CP_ACP, 0,
+			    win32font->logfontw.lfFaceName, -1,
+			    lfp->lfFaceName, G_N_ELEMENTS (lfp->lfFaceName),
+			    NULL, NULL))
+    lfp->lfFaceName[0] = '\0';
+
+  return lfp;
+}
+
+/**
+ * pango_win32_font_logfontw:
+ * @font: a #PangoFont which must be from the Win32 backend
+ * 
+ * Determine the LOGFONTW struct for the specified font.
+ * 
+ * Return value: A newly allocated LOGFONTW struct. It must be
+ * freed with g_free().
+ *
+ *Since: 1.16
+ **/
+LOGFONTW *
+pango_win32_font_logfontw (PangoFont *font)
+{
+  PangoWin32Font *win32font = (PangoWin32Font *)font;
+  LOGFONTW *lfp;
+
+  g_return_val_if_fail (font != NULL, NULL);
+  g_return_val_if_fail (PANGO_WIN32_IS_FONT (font), NULL);
+
+  lfp = g_new (LOGFONTW, 1);
+  *lfp = win32font->logfontw;
 
   return lfp;
 }
@@ -700,7 +719,7 @@ pango_win32_font_logfont (PangoFont *font)
  * pango_win32_font_select_font:
  * @font: a #PangoFont from the Win32 backend
  * @hdc: a windows device context
- * 
+ *
  * Selects the font into the specified DC and changes the mapping mode
  * and world transformation of the DC appropriately for the font.
  * You may want to surround the use of this function with calls
@@ -710,7 +729,7 @@ pango_win32_font_logfont (PangoFont *font)
  * See pango_win32_font_get_metrics_factor() for information about
  * converting from the coordinate space used by this function
  * into Pango units.
- * 
+ *
  * Return value: %TRUE if the operation succeeded.
  **/
 gboolean
@@ -718,7 +737,7 @@ pango_win32_font_select_font (PangoFont *font,
 			      HDC        hdc)
 {
   g_return_val_if_fail (PANGO_WIN32_IS_FONT (font), FALSE);
-  
+
   return PANGO_WIN32_FONT_GET_CLASS (font)->select_font (font, hdc);
 }
 
@@ -732,18 +751,18 @@ void
 pango_win32_font_done_font (PangoFont *font)
 {
   g_return_if_fail (PANGO_WIN32_IS_FONT (font));
-  
+
   PANGO_WIN32_FONT_GET_CLASS (font)->done_font (font);
 }
 
 /**
  * pango_win32_font_get_metrics_factor:
  * @font: a #PangoFont from the win32 backend
- * 
+ *
  * Returns the scale factor from logical units in the coordinate
  * space used by pango_win32_font_select_font() to Pango units
  * in user space.
- * 
+ *
  * Return value: factor to multiply logical units by to get Pango
  *               units.
  **/
@@ -751,10 +770,27 @@ double
 pango_win32_font_get_metrics_factor (PangoFont *font)
 {
   g_return_val_if_fail (PANGO_WIN32_IS_FONT (font), 1.);
-  
+
   return PANGO_WIN32_FONT_GET_CLASS (font)->get_metrics_factor (font);
 }
 
+static void
+pango_win32_fontmap_cache_add (PangoFontMap   *fontmap,
+			       PangoWin32Font *win32font)
+{
+  PangoWin32FontMap *win32fontmap = PANGO_WIN32_FONT_MAP (fontmap);
+
+  if (win32fontmap->freed_fonts->length == MAX_FREED_FONTS)
+    {
+      PangoWin32Font *old_font = g_queue_pop_tail (win32fontmap->freed_fonts);
+      g_object_unref (old_font);
+    }
+
+  g_object_ref (win32font);
+  g_queue_push_head (win32fontmap->freed_fonts, win32font);
+  win32font->in_cache = TRUE;
+}
+
 static void
 pango_win32_font_dispose (GObject *object)
 {
@@ -767,7 +803,7 @@ pango_win32_font_dispose (GObject *object)
   if (!win32font->in_cache && win32font->fontmap)
     pango_win32_fontmap_cache_add (win32font->fontmap, win32font);
 
-  G_OBJECT_CLASS (pango_win32_font_parent_class)->dispose (object);
+  G_OBJECT_CLASS (_pango_win32_font_parent_class)->dispose (object);
 }
 
 static void
@@ -777,6 +813,13 @@ free_metrics_info (PangoWin32MetricsInfo *info)
   g_free (info);
 }
 
+static void
+pango_win32_font_entry_remove (PangoWin32Face *face,
+			       PangoFont      *font)
+{
+  face->cached_fonts = g_slist_remove (face->cached_fonts, font);
+}
+
 static void
 pango_win32_font_finalize (GObject *object)
 {
@@ -788,15 +831,15 @@ pango_win32_font_finalize (GObject *object)
 
   g_slist_foreach (win32font->metrics_by_lang, (GFunc)free_metrics_info, NULL);
   g_slist_free (win32font->metrics_by_lang);
-  
+
   if (win32font->win32face)
     pango_win32_font_entry_remove (win32font->win32face, PANGO_FONT (win32font));
- 
+
   g_hash_table_destroy (win32font->glyph_info);
- 
+
   g_object_unref (win32font->fontmap);
 
-  G_OBJECT_CLASS (pango_win32_font_parent_class)->finalize (object);
+  G_OBJECT_CLASS (_pango_win32_font_parent_class)->finalize (object);
 }
 
 static PangoFontDescription *
@@ -807,7 +850,7 @@ pango_win32_font_describe (PangoFont *font)
 
   desc = pango_font_description_copy (win32font->win32face->description);
   pango_font_description_set_size (desc, win32font->size / (PANGO_SCALE / PANGO_WIN32_FONT_MAP (win32font->fontmap)->resolution));
-  
+
   return desc;
 }
 
@@ -819,25 +862,64 @@ pango_win32_font_describe_absolute (PangoFont *font)
 
   desc = pango_font_description_copy (win32font->win32face->description);
   pango_font_description_set_absolute_size (desc, win32font->size);
-  
+
   return desc;
 }
 
-PangoMap *
+static PangoMap *
 pango_win32_get_shaper_map (PangoLanguage *lang)
 {
   static guint engine_type_id = 0;
   static guint render_type_id = 0;
-  
+
   if (engine_type_id == 0)
     {
       engine_type_id = g_quark_from_static_string (PANGO_ENGINE_TYPE_SHAPE);
       render_type_id = g_quark_from_static_string (PANGO_RENDER_TYPE_WIN32);
     }
-  
+
   return pango_find_map (lang, engine_type_id, render_type_id);
 }
 
+static gint
+pango_win32_coverage_language_classify (PangoLanguage *lang)
+{
+  if (pango_language_matches (lang, "zh-tw"))
+    return PANGO_WIN32_COVERAGE_ZH_TW;
+  else if (pango_language_matches (lang, "zh-cn"))
+    return PANGO_WIN32_COVERAGE_ZH_CN;
+  else if (pango_language_matches (lang, "ja"))
+    return PANGO_WIN32_COVERAGE_JA;
+  else if (pango_language_matches (lang, "ko"))
+    return PANGO_WIN32_COVERAGE_KO;
+  else if (pango_language_matches (lang, "vi"))
+    return PANGO_WIN32_COVERAGE_VI;
+  else
+    return PANGO_WIN32_COVERAGE_UNSPEC;
+}
+
+static PangoCoverage *
+pango_win32_font_entry_get_coverage (PangoWin32Face *face,
+				     PangoLanguage  *lang)
+{
+  gint i = pango_win32_coverage_language_classify (lang);
+  if (face->coverages[i])
+    {
+      pango_coverage_ref (face->coverages[i]);
+      return face->coverages[i];
+    }
+
+  return NULL;
+}
+
+static void
+pango_win32_font_entry_set_coverage (PangoWin32Face *face,
+				     PangoCoverage  *coverage,
+				     PangoLanguage  *lang)
+{
+  face->coverages[pango_win32_coverage_language_classify (lang)] = pango_coverage_ref (coverage);
+}
+
 static PangoCoverage *
 pango_win32_font_get_coverage (PangoFont     *font,
 			       PangoLanguage *lang)
@@ -850,7 +932,7 @@ pango_win32_font_get_coverage (PangoFont     *font,
     {
       coverage = pango_coverage_new ();
       pango_win32_font_calc_coverage (font, coverage, lang);
-      
+
       pango_win32_font_entry_set_coverage (win32font->win32face, coverage, lang);
     }
 
@@ -876,12 +958,12 @@ pango_win32_font_find_shaper (PangoFont     *font,
  * pango_win32_get_unknown_glyph:
  * @font: a #PangoFont
  * @wc: the Unicode character for which a glyph is needed.
- * 
+ *
  * Returns the index of a glyph suitable for drawing @wc as an
  * unknown character.
  *
  * Use PANGO_GET_UNKNOWN_GLYPH() instead.
- * 
+ *
  * Return value: a glyph index into @font
  **/
 PangoGlyph
@@ -902,21 +984,21 @@ pango_win32_get_unknown_glyph (PangoFont *font,
  * work property the text alignment of the DC should have TA_BASELINE
  * and TA_LEFT.
  */
-void 
+void
 pango_win32_render_layout_line (HDC              hdc,
 				PangoLayoutLine *line,
-				int              x, 
+				int              x,
 				int              y)
 {
   GSList *tmp_list = line->runs;
   PangoRectangle overall_rect;
   PangoRectangle logical_rect;
   PangoRectangle ink_rect;
-  
+
   int x_off = 0;
 
   pango_layout_line_get_extents (line,NULL, &overall_rect);
-  
+
   while (tmp_list)
     {
       HBRUSH oldfg = NULL;
@@ -926,7 +1008,7 @@ pango_win32_render_layout_line (HDC              hdc,
       PangoLayoutRun *run = tmp_list->data;
       PangoAttrColor fg_color, bg_color;
       gboolean fg_set, bg_set;
-      
+
       tmp_list = tmp_list->next;
 
       pango_win32_get_item_properties (run->item, &uline, &fg_color, &fg_set, &bg_color, &bg_set);
@@ -984,27 +1066,27 @@ pango_win32_render_layout_line (HDC              hdc,
 	  Polyline (hdc, points, 2);
 	  break;
 	case PANGO_UNDERLINE_ERROR:
-          {
-            int point_x;
-            int counter = 0;
+	  {
+	    int point_x;
+	    int counter = 0;
 	    int end_x = x + PANGO_PIXELS (x_off + ink_rect.x + ink_rect.width);
 
-            for (point_x = x + PANGO_PIXELS (x_off + ink_rect.x) - 1;
-                 point_x <= end_x;
-                 point_x += 2)
-            {
+	    for (point_x = x + PANGO_PIXELS (x_off + ink_rect.x) - 1;
+		 point_x <= end_x;
+		 point_x += 2)
+	    {
 	      points[0].x = point_x;
 	      points[1].x = MAX (point_x + 1, end_x);
 
-              if (counter)
-	        points[0].y = points[1].y = y + 2;
-              else
-	        points[0].y = points[1].y = y + 3;
+	      if (counter)
+		points[0].y = points[1].y = y + 2;
+	      else
+		points[0].y = points[1].y = y + 3;
 
 	      Polyline (hdc, points, 2);
-              counter = (counter + 1) % 2;
-            }
-          }
+	      counter = (counter + 1) % 2;
+	    }
+	  }
 	  break;
 	case PANGO_UNDERLINE_LOW:
 	  points[0].x = x + PANGO_PIXELS (x_off + ink_rect.x) - 1;
@@ -1019,7 +1101,7 @@ pango_win32_render_layout_line (HDC              hdc,
 	  SelectObject (hdc, oldfg);
 	  DeleteObject (brush);
 	}
-      
+
       x_off += logical_rect.width;
     }
 }
@@ -1033,10 +1115,10 @@ pango_win32_render_layout_line (HDC              hdc,
  *
  * Render a #PangoLayoutLine onto an X drawable
  */
-void 
+void
 pango_win32_render_layout (HDC          hdc,
 			   PangoLayout *layout,
-			   int          x, 
+			   int          x,
 			   int          y)
 {
   PangoLayoutIter *iter;
@@ -1051,20 +1133,20 @@ pango_win32_render_layout (HDC          hdc,
       PangoRectangle   logical_rect;
       PangoLayoutLine *line;
       int              baseline;
-      
-      line = pango_layout_iter_get_line (iter);
-      
+
+      line = pango_layout_iter_get_line_readonly (iter);
+
       pango_layout_iter_get_line_extents (iter, NULL, &logical_rect);
       baseline = pango_layout_iter_get_baseline (iter);
-      
+
       pango_win32_render_layout_line (hdc,
-                                      line,
-                                      x + PANGO_PIXELS (logical_rect.x),
-                                      y + PANGO_PIXELS (baseline));
+				      line,
+				      x + PANGO_PIXELS (logical_rect.x),
+				      y + PANGO_PIXELS (baseline));
     }
   while (pango_layout_iter_next_line (iter));
 
-  pango_layout_iter_free (iter);  
+  pango_layout_iter_free (iter);
 }
 
 /* This utility function is duplicated here and in pango-layout.c; should it be
@@ -1082,10 +1164,10 @@ pango_win32_get_item_properties (PangoItem      *item,
 
   if (fg_set)
     *fg_set = FALSE;
-  
+
   if (bg_set)
     *bg_set = FALSE;
-  
+
   while (tmp_list)
     {
       PangoAttribute *attr = tmp_list->data;
@@ -1096,23 +1178,23 @@ pango_win32_get_item_properties (PangoItem      *item,
 	  if (uline)
 	    *uline = ((PangoAttrInt *)attr)->value;
 	  break;
-	  
+
 	case PANGO_ATTR_FOREGROUND:
 	  if (fg_color)
 	    *fg_color = *((PangoAttrColor *)attr);
 	  if (fg_set)
 	    *fg_set = TRUE;
-	  
+
 	  break;
-	  
+
 	case PANGO_ATTR_BACKGROUND:
 	  if (bg_color)
 	    *bg_color = *((PangoAttrColor *)attr);
 	  if (bg_set)
 	    *bg_set = TRUE;
-	  
+
 	  break;
-	  
+
 	default:
 	  break;
 	}
@@ -1120,7 +1202,7 @@ pango_win32_get_item_properties (PangoItem      *item,
     }
 }
 
-static guint 
+static guint
 get_cmap_offset (HDC     hdc,
 		 guint16 encoding_id)
 {
@@ -1136,9 +1218,9 @@ get_cmap_offset (HDC     hdc,
     return 0;
 
   n_tables = GUINT16_FROM_BE (n_tables);
-  
+
   table = g_malloc (ENCODING_TABLE_SIZE*n_tables);
-  
+
   res = GetFontData (hdc, CMAP, CMAP_HEADER_SIZE, table, ENCODING_TABLE_SIZE*n_tables);
   if (res != ENCODING_TABLE_SIZE*n_tables)
     return 0;
@@ -1166,8 +1248,8 @@ get_format_4_cmap (HDC hdc)
   guint16 *tbl, *tbl_end;
   struct format_4_cmap *table;
 
-  /* FIXME: Could look here at the CRC for the font in the DC 
-            and return a cached copy if the same */
+  /* FIXME: Could look here at the CRC for the font in the DC
+	    and return a cached copy if the same */
 
   offset = get_cmap_offset (hdc, UNICODE_ENCODING_ID);
   if (offset == 0)
@@ -1188,7 +1270,7 @@ get_format_4_cmap (HDC hdc)
       g_free (table);
       return NULL;
     }
-  
+
   table->format = GUINT16_FROM_BE (table->format);
   table->length = GUINT16_FROM_BE (table->length);
   table->language = GUINT16_FROM_BE (table->language);
@@ -1234,13 +1316,13 @@ static guint16 *
 get_end_count (struct format_4_cmap *table)
 {
   gint32 seg_count = table->seg_count_x_2/2;
-  /* Apparently the reseved spot is not reserved for 
+  /* Apparently the reseved spot is not reserved for
      the end_count array!? */
   return (&table->arrays[seg_count*0])-1;
 }
 
 static gboolean
-find_segment (struct format_4_cmap *table, 
+find_segment (struct format_4_cmap *table,
 	      guint16               wc,
 	      guint16              *segment)
 {
@@ -1249,7 +1331,7 @@ find_segment (struct format_4_cmap *table,
   guint16 *end_count = get_end_count (table);
   guint16 *start_count = get_start_count (table);
   static guint last = 0; /* Cache of one */
-  
+
   if (last < seg_count &&
       wc >= start_count[last] &&
       wc <= end_count[last])
@@ -1257,12 +1339,12 @@ find_segment (struct format_4_cmap *table,
       *segment = last;
       return TRUE;
     }
-      
+
 
   /* Binary search for the segment */
   start = 0; /* inclusive */
   end = seg_count; /* not inclusive */
-  while (start < end) 
+  while (start < end)
     {
       /* Look at middle pos */
       i = (start + end)/2;
@@ -1270,7 +1352,7 @@ find_segment (struct format_4_cmap *table,
       if (i == start)
 	{
 	  /* We made no progress. Look if this is the one. */
-	  
+
 	  if (wc >= start_count[i] &&
 	      wc <= end_count[i])
 	    {
@@ -1308,7 +1390,7 @@ get_format_12_cmap (HDC hdc)
   guint32 length;
   guint32 *tbl, *tbl_end;
   struct format_12_cmap *table;
-  
+
   offset = get_cmap_offset (hdc, UCS4_ENCODING_ID);
   if (offset == 0)
     return NULL;
@@ -1340,7 +1422,7 @@ get_format_12_cmap (HDC hdc)
       g_free (table);
       return NULL;
     }
-  
+
   tbl_end = (guint32 *) ((char *) table + length);
   tbl = table->groups;
 
@@ -1352,7 +1434,7 @@ get_format_12_cmap (HDC hdc)
 
   return table;
 }
-  
+
 static gpointer
 font_get_cmap (PangoFont *font)
 {
@@ -1362,15 +1444,15 @@ font_get_cmap (PangoFont *font)
   if (win32font->win32face->cmap)
     return win32font->win32face->cmap;
 
-  pango_win32_font_select_font (font, pango_win32_hdc);
+  pango_win32_font_select_font (font, _pango_win32_hdc);
 
   /* Prefer the format 12 cmap */
-  if ((cmap = get_format_12_cmap (pango_win32_hdc)) != NULL)
+  if ((cmap = get_format_12_cmap (_pango_win32_hdc)) != NULL)
     {
       win32font->win32face->cmap_format = 12;
       win32font->win32face->cmap = cmap;
     }
-  else if ((cmap = get_format_4_cmap (pango_win32_hdc)) != NULL)
+  else if ((cmap = get_format_4_cmap (_pango_win32_hdc)) != NULL)
     {
       win32font->win32face->cmap_format = 4;
       win32font->win32face->cmap = cmap;
@@ -1414,7 +1496,7 @@ pango_win32_font_get_glyph_index (PangoFont *font,
       guint16 segment;
       guint16 id;
       guint16 ch = wc;
-  
+
       if (wc > 0xFFFF)
 	return 0;
 
@@ -1460,8 +1542,8 @@ pango_win32_font_get_glyph_index (PangoFont *font,
 }
 
 gboolean
-pango_win32_get_name_header (HDC                 hdc,
-			     struct name_header *header)
+_pango_win32_get_name_header (HDC                 hdc,
+			      struct name_header *header)
 {
   if (GetFontData (hdc, NAME, 0, header, sizeof (*header)) != sizeof (*header))
     return FALSE;
@@ -1471,16 +1553,16 @@ pango_win32_get_name_header (HDC                 hdc,
 
   return TRUE;
 }
-  
+
 gboolean
-pango_win32_get_name_record (HDC                 hdc,
-			     gint                i,
-			     struct name_record *record)
+_pango_win32_get_name_record (HDC                 hdc,
+			      gint                i,
+			      struct name_record *record)
 {
   if (GetFontData (hdc, NAME, 6 + i * sizeof (*record),
 		   record, sizeof (*record)) != sizeof (*record))
     return FALSE;
-  
+
   record->platform_id = GUINT16_FROM_BE (record->platform_id);
   record->encoding_id = GUINT16_FROM_BE (record->encoding_id);
   record->language_id = GUINT16_FROM_BE (record->language_id);
@@ -1505,19 +1587,19 @@ font_has_name_in (PangoFont                       *font,
     return TRUE;
 
   hfont = pango_win32_get_hfont (font);
-  oldhfont = SelectObject (pango_win32_hdc, hfont);
+  oldhfont = SelectObject (_pango_win32_hdc, hfont);
 
-  if (!pango_win32_get_name_header (pango_win32_hdc, &header))
+  if (!_pango_win32_get_name_header (_pango_win32_hdc, &header))
     {
-      SelectObject (pango_win32_hdc, oldhfont);
+      SelectObject (_pango_win32_hdc, oldhfont);
       return FALSE;
     }
 
   for (i = 0; i < header.num_records; i++)
     {
-      if (!pango_win32_get_name_record (pango_win32_hdc, i, &record))
+      if (!_pango_win32_get_name_record (_pango_win32_hdc, i, &record))
 	{
-	  SelectObject (pango_win32_hdc, oldhfont);
+	  SelectObject (_pango_win32_hdc, oldhfont);
 	  return FALSE;
 	}
 
@@ -1549,7 +1631,7 @@ font_has_name_in (PangoFont                       *font,
 	  }
     }
 
-  SelectObject (pango_win32_hdc, oldhfont);
+  SelectObject (_pango_win32_hdc, oldhfont);
   return retval;
 }
 
@@ -1564,7 +1646,7 @@ pango_win32_font_calc_coverage (PangoFont     *font,
   guint32 i;
   PangoWin32CoverageLanguageClass cjkv;
   gboolean hide_unihan = FALSE;
-  
+
   PING(("font:%s lang:%s",
 	pango_font_description_to_string (pango_font_describe (font)),
 	pango_language_to_string (lang)));
@@ -1584,7 +1666,7 @@ pango_win32_font_calc_coverage (PangoFont     *font,
       PING(("no table"));
       return;
     }
-  
+
   PING (("coverage:"));
   if (win32font->win32face->cmap_format == 4)
     {
@@ -1605,14 +1687,14 @@ pango_win32_font_calc_coverage (PangoFont     *font,
 	  if (id_range_offset[i] == 0)
 	    {
 #ifdef PANGO_WIN32_DEBUGGING
-	      if (pango_win32_debug)
+	      if (_pango_win32_debug)
 		{
 		  if (end_count[i] == start_count[i])
 		    g_print ("%04x ", start_count[i]);
 		  else
 		    g_print ("%04x:%04x ", start_count[i], end_count[i]);
 		}
-#endif	  
+#endif
 	      for (ch = start_count[i]; ch <= end_count[i]; ch++)
 		if (hide_unihan && CH_IS_UNIHAN_BMP (ch))
 		  pango_coverage_set (coverage, ch, PANGO_COVERAGE_APPROXIMATE);
@@ -1628,7 +1710,7 @@ pango_win32_font_calc_coverage (PangoFont     *font,
 		{
 		  if (ch == 0xFFFF)
 		    break;
-		  
+
 		  id = *(id_range_offset[i]/2 +
 			 (ch - start_count[i]) +
 			 &id_range_offset[i]);
@@ -1646,7 +1728,7 @@ pango_win32_font_calc_coverage (PangoFont     *font,
 #ifdef PANGO_WIN32_DEBUGGING
 		  else if (ch0 < G_MAXUINT)
 		    {
-		      if (pango_win32_debug)
+		      if (_pango_win32_debug)
 			{
 			  if (ch > ch0 + 2)
 			    g_print ("%04x:%04x ", ch0, ch - 1);
@@ -1660,7 +1742,7 @@ pango_win32_font_calc_coverage (PangoFont     *font,
 #ifdef PANGO_WIN32_DEBUGGING
 	      if (ch0 < G_MAXUINT)
 		{
-		  if (pango_win32_debug)
+		  if (_pango_win32_debug)
 		    {
 		      if (ch > ch0 + 2)
 			g_print ("%04x:%04x ", ch0, ch - 1);
@@ -1679,7 +1761,7 @@ pango_win32_font_calc_coverage (PangoFont     *font,
       for (i = 0; i < cmap12->count; i++)
 	{
 #ifdef PANGO_WIN32_DEBUGGING
-	  if (pango_win32_debug)
+	  if (_pango_win32_debug)
 	    {
 	      if (cmap12->groups[i*3+0] == cmap12->groups[i*3+1])
 		g_print ("%04x ", cmap12->groups[i*3+0]);
@@ -1699,7 +1781,7 @@ pango_win32_font_calc_coverage (PangoFont     *font,
   else
     g_assert_not_reached ();
 #ifdef PANGO_WIN32_DEBUGGING
-  if (pango_win32_debug)
+  if (_pango_win32_debug)
     g_print ("\n");
 #endif
 }
